@@ -1,6 +1,6 @@
 /*******************************************************************************
 
-    uBlock Origin - a browser extension to block requests.
+    uBlock Origin - a comprehensive, efficient content blocker
     Copyright (C) 2014-present Raymond Hill
 
     This program is free software: you can redistribute it and/or modify
@@ -30,21 +30,15 @@
 
 /******************************************************************************/
 
-if ( typeof vAPI !== 'object' || vAPI === null ) {
-    return;
-}
+if ( typeof vAPI !== 'object' ) { return; }
+if ( typeof vAPI === null ) { return; }
 
-/******************************************************************************/
+if ( vAPI.pickerFrame ) { return; }
+vAPI.pickerFrame = true;
 
-const epickerId = vAPI.randomToken();
-let epickerConnectionId;
+const pickerUniqueId = vAPI.randomToken();
 
-let pickerRoot = document.querySelector(`[${vAPI.sessionId}]`);
-if ( pickerRoot !== null ) { return; }
-
-let pickerBootArgs;
-
-const reCosmeticAnchor = /^#[$?]?#/;
+const reCosmeticAnchor = /^#(\$|\?|\$\?)?#/;
 
 const netFilterCandidates = [];
 const cosmeticFilterCandidates = [];
@@ -129,7 +123,7 @@ const highlightElements = function(elems, force) {
     const islands = [];
 
     for ( const elem of elems ) {
-        if ( elem === pickerRoot ) { continue; }
+        if ( elem === pickerFrame ) { continue; }
         targetElements.push(elem);
         const rect = getElementBoundingClientRect(elem);
         // Ignore offscreen areas
@@ -144,7 +138,7 @@ const highlightElements = function(elems, force) {
         );
     }
 
-    vAPI.MessagingConnection.sendTo(epickerConnectionId, {
+    pickerFramePort.postMessage({
         what: 'svgPaths',
         ocean: `M0 0h${ow}v${oh}h-${ow}z`,
         islands: islands.join(''),
@@ -221,7 +215,8 @@ const backgroundImageURLFromElement = function(elem) {
 // https://github.com/gorhill/uBlock/issues/1725#issuecomment-226479197
 //   Limit returned string to 1024 characters.
 //   Also, return only URLs which will be seen by an HTTP observer.
-
+// https://github.com/uBlockOrigin/uBlock-issues/issues/2260
+//   Maybe get to the actual URL indirectly.
 const resourceURLsFromElement = function(elem) {
     const urls = [];
     const tagName = elem.localName;
@@ -231,7 +226,10 @@ const resourceURLsFromElement = function(elem) {
         if ( url !== '' ) { urls.push(url); }
         return urls;
     }
-    const s = elem[prop];
+    let s = elem[prop];
+    if ( s instanceof SVGAnimatedString ) {
+        s = s.baseVal;
+    }
     if ( typeof s === 'string' && /^https?:\/\//.test(s) ) {
         urls.push(trimFragmentFromURL(s.slice(0, 1024)));
     }
@@ -376,6 +374,7 @@ const netFilter1stSources = {
      'embed': 'src',
     'iframe': 'src',
        'img': 'src',
+     'image': 'href',
     'object': 'data',
     'source': 'src',
      'video': 'src'
@@ -550,10 +549,10 @@ const filtersFrom = function(x, y) {
     // https://www.reddit.com/r/uBlockOrigin/comments/qmjk36/
     //   Extract network candidates first.
     if ( typeof x === 'number' ) {
-        const magicAttr = `${vAPI.sessionId}-clickblind`;
-        pickerRoot.setAttribute(magicAttr, '');
+        const magicAttr = `${pickerUniqueId}-clickblind`;
+        pickerFrame.setAttribute(magicAttr, '');
         const elems = document.elementsFromPoint(x, y);
-        pickerRoot.removeAttribute(magicAttr);
+        pickerFrame.removeAttribute(magicAttr);
         for ( const elem of elems ) {
             netFilterFromElement(elem);
         }
@@ -620,6 +619,21 @@ const filterToDOMInterface = (( ) => {
     const reCaret = '(?:[^%.0-9a-z_-]|$)';
     const rePseudoElements = /:(?::?after|:?before|:[a-z-]+)$/;
 
+    const matchElemToRegex = (elem, re) => {
+        const srcProp = netFilter1stSources[elem.localName];
+        let src = elem[srcProp];
+        if ( src instanceof SVGAnimatedString ) {
+            src = src.baseVal;
+        }
+        if ( typeof src === 'string' && /^https?:\/\//.test(src) ) {
+            if ( re.test(src) ) { return srcProp; }
+        }
+        src = elem.currentSrc;
+        if ( typeof src === 'string' && /^https?:\/\//.test(src) ) {
+            if ( re.test(src) ) { return srcProp; }
+        }
+    };
+
     // Net filters: we need to lookup manually -- translating into a foolproof
     // CSS selector is just not possible.
     //
@@ -671,25 +685,23 @@ const filterToDOMInterface = (( ) => {
         }
 
         // Lookup by tag names.
+        // https://github.com/uBlockOrigin/uBlock-issues/issues/2260
+        //   Maybe get to the actual URL indirectly.
+        //
+        // https://github.com/uBlockOrigin/uBlock-issues/issues/3142
+        //   Don't try to match against non-network URIs.
         const elems = document.querySelectorAll(
             Object.keys(netFilter1stSources).join()
         );
         for ( const elem of elems ) {
-            const srcProp = netFilter1stSources[elem.localName];
-            const src = elem[srcProp];
-            if (
-                typeof src === 'string' &&
-                    reFilter.test(src) ||
-                typeof elem.currentSrc === 'string' &&
-                    reFilter.test(elem.currentSrc)
-            ) {
-                out.push({
-                    elem,
-                    src: srcProp,
-                    opt: filterTypes[elem.localName],
-                    style: vAPI.hideStyle,
-                });
-            }
+            const srcProp = matchElemToRegex(elem, reFilter);
+            if ( srcProp === undefined ) { continue; }
+            out.push({
+                elem,
+                src: srcProp,
+                opt: filterTypes[elem.localName],
+                style: vAPI.hideStyle,
+            });
         }
 
         // Find matching background image in current set of candidate elements.
@@ -728,7 +740,7 @@ const filterToDOMInterface = (( ) => {
         }
         const out = [];
         for ( const elem of elems ) {
-            if ( elem === pickerRoot ) { continue; }
+            if ( elem === pickerFrame ) { continue; }
             out.push({ elem, raw, style: vAPI.hideStyle });
         }
         return out;
@@ -746,9 +758,17 @@ const filterToDOMInterface = (( ) => {
         try {
             const o = JSON.parse(raw);
             elems = vAPI.domFilterer.createProceduralFilter(o).exec();
-            style = o.action === undefined || o.action[0] !== ':style'
-                ? vAPI.hideStyle
-                : o.action[1];
+            switch ( o.action && o.action[0] || '' ) {
+            case '':
+            case 'remove':
+                style = vAPI.hideStyle;
+                break;
+            case 'style':
+                style = o.action[1];
+                break;
+            default:
+                break;
+            }
         } catch(ex) {
             return;
         }
@@ -798,7 +818,8 @@ const filterToDOMInterface = (( ) => {
         if ( Array.isArray(lastResultset) === false ) { return; }
         const rootElem = document.documentElement;
         for ( const { elem, style } of lastResultset ) {
-            if ( elem === pickerRoot ) { continue; }
+            if ( elem === pickerFrame ) { continue; }
+            if ( style === undefined ) { continue; }
             if ( elem === rootElem && style === vAPI.hideStyle ) { continue; }
             let styleToken = vAPI.epickerStyleProxies.get(style);
             if ( styleToken === undefined ) {
@@ -858,7 +879,7 @@ const filterToDOMInterface = (( ) => {
 
 /******************************************************************************/
 
-const onOptmizeCandidates = function(details) {
+const onOptimizeCandidates = function(details) {
     const { candidates } = details;
     const results = [];
     for ( const paths of candidates ) {
@@ -881,7 +902,7 @@ const onOptmizeCandidates = function(details) {
         if ( r !== 0 ) { return r; }
         return a.selector.length - b.selector.length;
     });
-    vAPI.MessagingConnection.sendTo(epickerConnectionId, {
+    pickerFramePort.postMessage({
         what: 'candidatesOptimized',
         candidates: results.map(a => a.selector),
         slot: details.slot,
@@ -891,7 +912,7 @@ const onOptmizeCandidates = function(details) {
 /******************************************************************************/
 
 const showDialog = function(options) {
-    vAPI.MessagingConnection.sendTo(epickerConnectionId, {
+    pickerFramePort.postMessage({
         what: 'showDialog',
         url: self.location.href,
         netFilters: netFilterCandidates,
@@ -914,9 +935,9 @@ const elementFromPoint = (( ) => {
         } else {
             return null;
         }
-        if ( !pickerRoot ) { return null; }
-        const magicAttr = `${vAPI.sessionId}-clickblind`;
-        pickerRoot.setAttribute(magicAttr, '');
+        if ( !pickerFrame ) { return null; }
+        const magicAttr = `${pickerUniqueId}-clickblind`;
+        pickerFrame.setAttribute(magicAttr, '');
         let elem = document.elementFromPoint(x, y);
         if (
             elem === null || /* to skip following tests */
@@ -930,7 +951,7 @@ const elementFromPoint = (( ) => {
             elem = null;
         }
         // https://github.com/uBlockOrigin/uBlock-issues/issues/380
-        pickerRoot.removeAttribute(magicAttr);
+        pickerFrame.removeAttribute(magicAttr);
         return elem;
     };
 })();
@@ -1046,7 +1067,7 @@ const onViewportChanged = function() {
 // Auto-select a specific target, if any, and if possible
 
 const startPicker = function() {
-    pickerRoot.focus();
+    pickerFrame.focus();
 
     self.addEventListener('scroll', onViewportChanged, { passive: true });
     self.addEventListener('resize', onViewportChanged, { passive: true });
@@ -1083,7 +1104,7 @@ const startPicker = function() {
     if ( attr === undefined ) { return; }
     const elems = document.getElementsByTagName(tagName);
     for ( const elem of elems  ) {
-        if ( elem === pickerRoot ) { continue; }
+        if ( elem === pickerFrame ) { continue; }
         const srcs = resourceURLsFromElement(elem);
         if (
             (srcs.length !== 0 && srcs.includes(url) === false) ||
@@ -1122,31 +1143,35 @@ const quitPicker = function() {
     self.removeEventListener('resize', onViewportChanged, { passive: true });
     self.removeEventListener('keydown', onKeyPressed, true);
     vAPI.shutdown.remove(quitPicker);
-    vAPI.MessagingConnection.disconnectFrom(epickerConnectionId);
-    vAPI.MessagingConnection.removeListener(onConnectionMessage);
+    if ( pickerFramePort ) {
+        pickerFramePort.close();
+        pickerFramePort = null;
+    }
+    if ( pickerFrame ) {
+        pickerFrame.remove();
+        pickerFrame = null;
+    }
     vAPI.userStylesheet.remove(pickerCSS);
     vAPI.userStylesheet.apply();
-
-    if ( pickerRoot === null ) { return; }
-
-    pickerRoot.remove();
-    pickerRoot = null;
-
+    vAPI.pickerFrame = false;
     self.focus();
 };
 
+vAPI.shutdown.add(quitPicker);
+
 /******************************************************************************/
 
 const onDialogMessage = function(msg) {
     switch ( msg.what ) {
         case 'start':
             startPicker();
+            if ( pickerFramePort === null ) { break; }
             if ( targetElements.length === 0 ) {
                 highlightElements([], true);
             }
             break;
         case 'optimizeCandidates':
-            onOptmizeCandidates(msg);
+            onOptimizeCandidates(msg);
             break;
         case 'dialogCreate':
             filterToDOMInterface.queryAll(msg);
@@ -1157,7 +1182,7 @@ const onDialogMessage = function(msg) {
             const resultset = filterToDOMInterface.queryAll(msg) || [];
             highlightElements(resultset.map(a => a.elem), true);
             if ( msg.filter === '!' ) { break; }
-            vAPI.MessagingConnection.sendTo(epickerConnectionId, {
+            pickerFramePort.postMessage({
                 what: 'resultsetDetails',
                 count: resultset.length,
                 opt: resultset.length !== 0 ? resultset[0].opt : undefined,
@@ -1196,23 +1221,6 @@ const onDialogMessage = function(msg) {
 
 /******************************************************************************/
 
-const onConnectionMessage = function(msg) {
-    if ( msg.from !== `epickerDialog-${epickerId}` ) { return; }
-    switch ( msg.what ) {
-        case 'connectionRequested':
-            epickerConnectionId = msg.id;
-            return true;
-        case 'connectionBroken':
-            quitPicker();
-            break;
-        case 'connectionMessage':
-            onDialogMessage(msg.payload);
-            break;
-    }
-};
-
-/******************************************************************************/
-
 // epicker-ui.html will be injected in the page through an iframe, and
 // is a sandboxed so as to prevent the page from interfering with its
 // content and behavior.
@@ -1230,25 +1238,6 @@ const onConnectionMessage = function(msg) {
 // of the iframe, and cannot interfere with its style properties. However the
 // page can remove the iframe.
 
-// We need extra messaging capabilities + fetch/process picker arguments.
-{
-    const results = await Promise.all([
-        vAPI.messaging.extend(),
-        vAPI.messaging.send('elementPicker', { what: 'elementPickerArguments' }),
-    ]);
-    if ( results[0] !== true ) { return; }
-    pickerBootArgs = results[1];
-    if ( typeof pickerBootArgs !== 'object' || pickerBootArgs === null ) {
-        return;
-    }
-    // Restore net filter union data if origin is the same.
-    const eprom = pickerBootArgs.eprom || null;
-    if ( eprom !== null && eprom.lastNetFilterSession === lastNetFilterSession ) {
-        lastNetFilterHostname = eprom.lastNetFilterHostname || '';
-        lastNetFilterUnion = eprom.lastNetFilterUnion || '';
-    }
-}
-
 // The DOM filterer will not be present when cosmetic filtering is disabled.
 const noCosmeticFiltering =
     vAPI.domFilterer instanceof Object === false ||
@@ -1262,9 +1251,11 @@ const pickerCSSStyle = [
     'border: 0',
     'border-radius: 0',
     'box-shadow: none',
+    'color-scheme: light dark',
     'display: block',
     'filter: none',
     'height: 100vh',
+    '    height: 100svh',
     'left: 0',
     'margin: 0',
     'max-height: none',
@@ -1278,19 +1269,21 @@ const pickerCSSStyle = [
     'position: fixed',
     'top: 0',
     'transform: none',
-    'visibility: visible',
+    'visibility: hidden',
     'width: 100%',
     'z-index: 2147483647',
-    // https://github.com/uBlockOrigin/uBlock-issues/issues/1408
-    'color-scheme: light',
     ''
-];
+].join(' !important;\n');
+
 
 const pickerCSS = `
-:root > [${vAPI.sessionId}] {
-    ${pickerCSSStyle.join(' !important;')}
+:root > [${pickerUniqueId}] {
+    ${pickerCSSStyle}
 }
-:root [${vAPI.sessionId}-clickblind] {
+:root > [${pickerUniqueId}-loaded] {
+    visibility: visible !important;
+}
+:root [${pickerUniqueId}-clickblind] {
     pointer-events: none !important;
 }
 `;
@@ -1298,21 +1291,53 @@ const pickerCSS = `
 vAPI.userStylesheet.add(pickerCSS);
 vAPI.userStylesheet.apply();
 
-pickerRoot = document.createElement('iframe');
-pickerRoot.setAttribute(vAPI.sessionId, '');
-document.documentElement.append(pickerRoot);
+let pickerBootArgs;
+let pickerFramePort = null;
 
-vAPI.shutdown.add(quitPicker);
-
-vAPI.MessagingConnection.addListener(onConnectionMessage);
-
-{
+const bootstrap = async ( ) => {
+    pickerBootArgs = await vAPI.messaging.send('elementPicker', {
+        what: 'elementPickerArguments',
+    });
+    if ( typeof pickerBootArgs !== 'object' ) { return; }
+    if ( pickerBootArgs === null ) { return; }
+    // Restore net filter union data if origin is the same.
+    const eprom = pickerBootArgs.eprom || null;
+    if ( eprom !== null && eprom.lastNetFilterSession === lastNetFilterSession ) {
+        lastNetFilterHostname = eprom.lastNetFilterHostname || '';
+        lastNetFilterUnion = eprom.lastNetFilterUnion || '';
+    }
     const url = new URL(pickerBootArgs.pickerURL);
-    url.searchParams.set('epid', epickerId);
     if ( pickerBootArgs.zap ) {
         url.searchParams.set('zap', '1');
     }
-    pickerRoot.contentWindow.location = url.href;
+    return new Promise(resolve => {
+        const iframe = document.createElement('iframe');
+        iframe.setAttribute(pickerUniqueId, '');
+        document.documentElement.append(iframe);
+        iframe.addEventListener('load', ( ) => {
+            iframe.setAttribute(`${pickerUniqueId}-loaded`, '');
+            const channel = new MessageChannel();
+            pickerFramePort = channel.port1;
+            pickerFramePort.onmessage = ev => {
+                onDialogMessage(ev.data || {});
+            };
+            pickerFramePort.onmessageerror = ( ) => {
+                quitPicker();
+            };
+            iframe.contentWindow.postMessage(
+                { what: 'epickerStart' },
+                url.href,
+                [ channel.port2 ]
+            );
+            resolve(iframe);
+        }, { once: true });
+        iframe.contentWindow.location = url.href;
+    });
+};
+
+let pickerFrame = await bootstrap();
+if ( Boolean(pickerFrame) === false ) {
+    quitPicker();
 }
 
 /******************************************************************************/
